__all__ = ["cumulative_sum", "max", "mean", "min", "prod", "std", "sum", "var"]


from jax.experimental import array_api
from jaxtyping import ArrayLike
from quax import Value

from quaxed._types import DType
from quaxed._utils import quaxify


@quaxify
def cumulative_sum(
    x: ArrayLike,
    /,
    *,
    axis: int | None = None,
    dtype: DType | None = None,
    include_initial: bool = False,
) -> Value:
    return array_api.cumulative_sum(
        x, axis=axis, dtype=dtype, include_initial=include_initial
    )


@quaxify
def max(  # pylint: disable=redefined-builtin
    x: ArrayLike,
    /,
    *,
    axis: int | tuple[int, ...] | None = None,
    keepdims: bool = False,
) -> Value:
    return array_api.max(x, axis=axis, keepdims=keepdims)


@quaxify
def mean(
    x: ArrayLike,
    /,
    *,
    axis: int | tuple[int, ...] | None = None,
    keepdims: bool = False,
) -> Value:
    return array_api.mean(x, axis=axis, keepdims=keepdims)


@quaxify
def min(  # pylint: disable=redefined-builtin
    x: ArrayLike,
    /,
    *,
    axis: int | tuple[int, ...] | None = None,
    keepdims: bool = False,
) -> Value:
    return array_api.min(x, axis=axis, keepdims=keepdims)


@quaxify
def prod(
    x: ArrayLike,
    /,
    *,
    axis: int | tuple[int, ...] | None = None,
    dtype: DType | None = None,
    keepdims: bool = False,
) -> Value:
    return array_api.prod(x, axis=axis, dtype=dtype, keepdims=keepdims)


@quaxify
def std(
    x: ArrayLike,
    /,
    *,
    axis: int | tuple[int, ...] | None = None,
    correction: int | float = 0.0,
    keepdims: bool = False,
) -> Value:
    return array_api.std(x, axis=axis, correction=correction, keepdims=keepdims)


@quaxify
def sum(  # pylint: disable=redefined-builtin
    x: ArrayLike,
    /,
    *,
    axis: int | tuple[int, ...] | None = None,
    dtype: DType | None = None,
    keepdims: bool = False,
) -> Value:
    return array_api.sum(x, axis=axis, dtype=dtype, keepdims=keepdims)


@quaxify
def var(
    x: ArrayLike,
    /,
    *,
    axis: int | tuple[int, ...] | None = None,
    correction: int | float = 0.0,
    keepdims: bool = False,
) -> Value:
    return array_api.var(x, axis=axis, correction=correction, keepdims=keepdims)



"""Utility functions for quaxed."""

from typing import TypeVar

import quax

T = TypeVar("T")


def quaxify(func: T, *, filter_spec: bool | tuple[bool, ...] = True) -> T:
    """Quaxify, but makes mypy happy."""
    return quax.quaxify(func, filter_spec=filter_spec)



"""FFT functions."""

__all__ = [
    "fft",
    "ifft",
    "fftn",
    "ifftn",
    "rfft",
    "irfft",
    "rfftn",
    "irfftn",
    "hfft",
    "ihfft",
    "fftfreq",
    "rfftfreq",
    "fftshift",
    "ifftshift",
]


from jax.numpy import fft as _jax_fft

from quaxed._utils import quaxify

fft = quaxify(_jax_fft.fft)
ifft = quaxify(_jax_fft.ifft)
fftn = quaxify(_jax_fft.fftn)
ifftn = quaxify(_jax_fft.ifftn)
rfft = quaxify(_jax_fft.rfft)
irfft = quaxify(_jax_fft.irfft)
rfftn = quaxify(_jax_fft.rfftn)
irfftn = quaxify(_jax_fft.irfftn)
hfft = quaxify(_jax_fft.hfft)
ihfft = quaxify(_jax_fft.ihfft)
fftfreq = quaxify(_jax_fft.fftfreq)
rfftfreq = quaxify(_jax_fft.rfftfreq)
fftshift = quaxify(_jax_fft.fftshift)
ifftshift = quaxify(_jax_fft.ifftshift)



"""Test with JAX inputs."""

import jax.numpy as jnp
import pytest
from jax.experimental import array_api as jax_xp

import quaxed.array_api as xp

# =============================================================================
# Constants


def test_e():
    """Test `e`."""
    assert xp.e is jax_xp.e


def test_inf():
    """Test `inf`."""
    assert xp.inf is jax_xp.inf


def test_nan():
    """Test `nan`."""
    assert xp.nan is jax_xp.nan


def test_newaxis():
    """Test `newaxis`."""
    assert xp.newaxis is jax_xp.newaxis


def test_pi():
    """Test `pi`."""
    assert xp.pi is jax_xp.pi


# =============================================================================
# Creation functions


@pytest.mark.parametrize(
    ("start", "stop", "step"),
    [
        # int
        (3, None, 1),
        (3, 1, 1),
        (4, None, 2),
        (3, 1, 2),
        # float
        (3.0, None, 1),
        (3.0, 1.0, 1),
        (4.0, None, 2.0),
        (3.0, 1.0, 2.0),
        # Array
        (jnp.array(3), None, 1),
        (jnp.array(3), jnp.array(1), 1),
        (jnp.array(4), None, jnp.array(2)),
        (jnp.array(3), jnp.array(1), jnp.array(2)),
        # TODO: mixed
    ],
)
def test_arange(start, stop, step):
    """Test `arange`."""
    # TODO: test the start, stop, step, dtype, device arguments
    got = xp.arange(start, stop, step)
    expected = jax_xp.arange(start, stop, step)

    assert isinstance(got, jnp.ndarray)
    assert jnp.array_equal(got, expected)


# def test_asarray():
#     """Test `asarray`."""
#     got = xp.asarray([1, 2, 3])
#     expected = jax_xp.asarray([1, 2, 3])

#     assert isinstance(got, jnp.ndarray)
#     assert jnp.array_equal(got, expected)



from jax.lax.linalg import cholesky as cholesky
from jax.lax.linalg import eig as eig
from jax.lax.linalg import eigh as eigh
from jax.lax.linalg import hessenberg as hessenberg
from jax.lax.linalg import householder_product as householder_product
from jax.lax.linalg import lu as lu
from jax.lax.linalg import qdwh as qdwh
from jax.lax.linalg import qr as qr
from jax.lax.linalg import shur as shur
from jax.lax.linalg import svd as svd
from jax.lax.linalg import triangular_solve as triangular_solve
from jax.lax.linalg import tridiagonal as tridiagonal
from jax.lax.linalg import tridiagonal_solve as tridiagonal_solve



"""Quaxed :mod:`jax.numpy`."""
# pylint: disable=redefined-builtin

from typing import Any

from jaxtyping import install_import_hook

from . import _core, _creation_functions, _higher_order
from ._creation_functions import *
from ._higher_order import *

__all__: list[str] = []
__all__ += _core.__all__
__all__ += _higher_order.__all__
__all__ += _creation_functions.__all__


# TODO: consolidate with ``_core.__getattr__``.
def __getattr__(name: str) -> Any:
    if name in __all__:
        return getattr(_core, name)

    msg = f"module {__name__!r} has no attribute {name!r}"
    raise AttributeError(msg)


# TODO: figure out how to install this import hook, with the __getattr__.
install_import_hook("quaxed.numpy", None)



"""Test with JAX inputs."""

import jax.numpy as jnp
import pytest
from jax import lax

import quaxed.lax as qlax


def test_dir():
    """Test the `__dir__` method."""
    assert set(qlax.__dir__()) == set(qlax.__all__)


def test_linalg_dir():
    """Test the `__dir__` method."""
    assert set(qlax.linalg.__dir__()) == set(qlax.linalg.__all__)


def test_not_in_lax():
    with pytest.raises(AttributeError, match="Cannot get"):
        _ = qlax.for_sure_not_in_lax


def test_not_in_lax_linalg():
    with pytest.raises(AttributeError, match="Cannot get"):
        _ = qlax.linalg.for_sure_not_in_lax_linalg


# ==============================================================================
# Operators


def test_abs():
    """Test `quaxed.lax.abs`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.abs(x)
    exp = lax.abs(x)

    assert jnp.array_equal(got, exp)


def test_acos():
    """Test `quaxed.lax.acos`."""
    x = jnp.array([[0.1, 0.2], [0.3, 0.4]], dtype=float)
    got = qlax.acos(x)
    exp = lax.acos(x)

    assert jnp.array_equal(got, exp)


def test_acosh():
    """Test `quaxed.lax.acosh`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.acosh(x)
    exp = lax.acosh(x)

    assert jnp.array_equal(got, exp)


def test_add():
    """Test `quaxed.lax.add`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    y = jnp.array([[5, 6], [7, 8]], dtype=float)
    got = qlax.add(x, y)
    exp = lax.add(x, y)

    assert jnp.array_equal(got, exp)


@pytest.mark.skip(reason="Not implemented.")
def test_after_all():
    pass


def test_approx_max_k():
    """Test `quaxed.lax.approx_max_k`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.approx_max_k(x, 2)
    exp = lax.approx_max_k(x, 2)

    assert jnp.array_equal(got, exp)


def test_approx_min_k():
    """Test `quaxed.lax.approx_min_k`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.approx_min_k(x, 2)
    exp = lax.approx_min_k(x, 2)

    assert jnp.array_equal(got, exp)


def test_argmax():
    """Test `quaxed.lax.argmax`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.argmax(x, axis=0, index_dtype=int)
    exp = lax.argmax(x, axis=0, index_dtype=int)

    assert jnp.array_equal(got, exp)


def test_argmin():
    """Test `quaxed.lax.argmin`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.argmin(x, axis=0, index_dtype=int)
    exp = lax.argmin(x, axis=0, index_dtype=int)

    assert jnp.array_equal(got, exp)


def test_asin():
    """Test `quaxed.lax.asin`."""
    x = jnp.array([[0.1, 0.2], [0.3, 0.4]], dtype=float)
    got = qlax.asin(x)
    exp = lax.asin(x)

    assert jnp.array_equal(got, exp)


def test_asinh():
    """Test `quaxed.lax.asinh`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.asinh(x)
    exp = lax.asinh(x)

    assert jnp.array_equal(got, exp)


def test_atan():
    """Test `quaxed.lax.atan`."""
    x = jnp.array([[0.1, 0.2], [0.3, 0.4]], dtype=float)
    got = qlax.atan(x)
    exp = lax.atan(x)

    assert jnp.array_equal(got, exp)


def test_atan2():
    """Test `quaxed.lax.atan2`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    y = jnp.array([[5, 6], [7, 8]], dtype=float)
    got = qlax.atan2(x, y)
    exp = lax.atan2(x, y)

    assert jnp.array_equal(got, exp)


def test_atanh():
    """Test `quaxed.lax.atanh`."""
    x = jnp.array([[0.1, 0.2], [0.3, 0.4]], dtype=float)
    got = qlax.atanh(x)
    exp = lax.atanh(x)

    assert jnp.array_equal(got, exp)


def test_batch_matmul():
    """Test `quaxed.lax.batch_matmul`."""
    x = jnp.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]], dtype=float)
    y = jnp.array([[[9, 10], [11, 12]], [[13, 14], [15, 16]]], dtype=float)
    got = qlax.batch_matmul(x, y)
    exp = lax.batch_matmul(x, y)

    assert jnp.array_equal(got, exp)


def test_bessel_i0e():
    """Test `quaxed.lax.bessel_i0e`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.bessel_i0e(x)
    exp = lax.bessel_i0e(x)

    assert jnp.array_equal(got, exp)


def test_bessel_i1e():
    """Test `quaxed.lax.bessel_i1e`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.bessel_i1e(x)
    exp = lax.bessel_i1e(x)

    assert jnp.array_equal(got, exp)


def test_betainc():
    """Test `quaxed.lax.betainc`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float) / 10
    y = jnp.array([[5, 6], [7, 8]], dtype=float) / 10
    got = qlax.betainc(1.0, x, y)
    exp = lax.betainc(1.0, x, y)

    assert jnp.array_equal(got, exp)


def test_bitcast_convert_type():
    """Test `quaxed.lax.bitcast_convert_type`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.bitcast_convert_type(x, jnp.int32)
    exp = lax.bitcast_convert_type(x, jnp.int32)

    assert jnp.array_equal(got, exp)


def test_bitwise_and():
    """Test `quaxed.lax.bitwise_and`."""
    x = jnp.array([[1, 0], [0, 1]], dtype=int)
    got = qlax.bitwise_and(x, x)
    exp = lax.bitwise_and(x, x)

    assert jnp.array_equal(got, exp)


def test_bitwise_not():
    """Test `quaxed.lax.bitwise_not`."""
    x = jnp.array([[1, 0], [0, 1]], dtype=int)
    got = qlax.bitwise_not(x)
    exp = lax.bitwise_not(x)

    assert jnp.array_equal(got, exp)


def test_bitwise_or():
    """Test `quaxed.lax.bitwise_or`."""
    x = jnp.array([[1, 0], [0, 1]], dtype=int)
    got = qlax.bitwise_or(x, x)
    exp = lax.bitwise_or(x, x)

    assert jnp.array_equal(got, exp)


def test_bitwise_xor():
    """Test `quaxed.lax.bitwise_xor`."""
    x = jnp.array([[1, 0], [0, 1]], dtype=int)
    got = qlax.bitwise_xor(x, x)
    exp = lax.bitwise_xor(x, x)

    assert jnp.array_equal(got, exp)


def test_population_count():
    """Test `quaxed.lax.population_count`."""
    x = jnp.array([[1, 0], [0, 1]], dtype=int)
    got = qlax.population_count(x)
    exp = lax.population_count(x)

    assert jnp.array_equal(got, exp)


def test_broadcast():
    """Test `quaxed.lax.broadcast`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.broadcast(x, (1, 1))
    exp = lax.broadcast(x, (1, 1))

    assert jnp.array_equal(got, exp)


def test_broadcast_in_dim():
    """Test `quaxed.lax.broadcast_in_dim`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.broadcast_in_dim(x, (1, 1, 2, 2), (2, 3))
    exp = lax.broadcast_in_dim(x, (1, 1, 2, 2), (2, 3))

    assert jnp.array_equal(got, exp)


def test_broadcast_shapes():
    """Test `quaxed.lax.broadcast_shapes`."""
    x = (2, 3)
    y = (1, 3)
    got = qlax.broadcast_shapes(x, y)
    exp = lax.broadcast_shapes(x, y)

    assert got == exp


def test_broadcast_to_rank():
    """Test `quaxed.lax.broadcast_to_rank`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.broadcast_to_rank(x, 3)
    exp = lax.broadcast_to_rank(x, 3)

    assert jnp.array_equal(got, exp)


@pytest.mark.skip(reason="Not implemented.")
def test_broadcasted_iota():
    pass


def test_cbrt():
    """Test `quaxed.lax.cbrt`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.cbrt(x)
    exp = lax.cbrt(x)

    assert jnp.array_equal(got, exp)


def test_ceil():
    """Test `quaxed.lax.ceil`."""
    x = jnp.array([[1.1, 2.2], [3.3, 4.4]], dtype=float)
    got = qlax.ceil(x)
    exp = lax.ceil(x)

    assert jnp.array_equal(got, exp)


def test_clamp():
    """Test `quaxed.lax.clamp`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.clamp(2.0, x, 3.0)
    exp = lax.clamp(2.0, x, 3.0)

    assert jnp.array_equal(got, exp)


def test_clz():
    """Test `quaxed.lax.clz`."""
    x = jnp.array([[0, 2], [0, 4]], dtype=int)
    got = qlax.clz(x)
    exp = lax.clz(x)

    assert jnp.array_equal(got, exp)


def test_collapse():
    """Test `quaxed.lax.collapse`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.collapse(x, 1)
    exp = lax.collapse(x, 1)

    assert jnp.array_equal(got, exp)


def test_concatenate():
    """Test `quaxed.lax.concatenate`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    y = jnp.array([[5, 6], [7, 8]], dtype=float)
    got = qlax.concatenate((x, y), 0)
    exp = lax.concatenate((x, y), 0)

    assert jnp.array_equal(got, exp)


def test_conj():
    """Test `quaxed.lax.conj`."""
    x = jnp.array([[1 + 1j, 2 + 2j], [3 + 3j, 4 + 4j]], dtype=complex)
    got = qlax.conj(x)
    exp = lax.conj(x)

    assert jnp.array_equal(got, exp)


@pytest.mark.skip(reason="TODO: implement.")
def test_conv():
    """Test `quaxed.lax.conv`."""


def test_convert_element_type():
    """Test `quaxed.lax.convert_element_type`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.convert_element_type(x, jnp.int32)
    exp = lax.convert_element_type(x, jnp.int32)

    assert jnp.array_equal(got, exp)


@pytest.mark.skip(reason="TODO: implement.")
def test_conv_dimension_numbers():
    """Test `quaxed.lax.conv_dimension_numbers`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_conv_general_dilated():
    """Test `quaxed.lax.conv_general_dilated`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_conv_general_dilated_local():
    """Test `quaxed.lax.conv_general_dilated_local`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_conv_general_dilated_patches():
    """Test `quaxed.lax.conv_general_dilated_patches`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_conv_transpose():
    """Test `quaxed.lax.conv_transpose`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_conv_with_general_padding():
    """Test `quaxed.lax.conv_with_general_padding`."""


def test_cos():
    """Test `quaxed.lax.cos`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.cos(x)
    exp = lax.cos(x)

    assert jnp.array_equal(got, exp)


def test_cosh():
    """Test `quaxed.lax.cosh`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.cosh(x)
    exp = lax.cosh(x)

    assert jnp.array_equal(got, exp)


def test_cumlogsumexp():
    """Test `quaxed.lax.cumlogsumexp`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.cumlogsumexp(x, axis=0)
    exp = lax.cumlogsumexp(x, axis=0)

    assert jnp.array_equal(got, exp)


def test_cummax():
    """Test `quaxed.lax.cummax`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.cummax(x, axis=0)
    exp = lax.cummax(x, axis=0)

    assert jnp.array_equal(got, exp)


def test_cummin():
    """Test `quaxed.lax.cummin`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.cummin(x, axis=0)
    exp = lax.cummin(x, axis=0)

    assert jnp.array_equal(got, exp)


def test_cumprod():
    """Test `quaxed.lax.cumprod`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.cumprod(x, axis=0)
    exp = lax.cumprod(x, axis=0)

    assert jnp.array_equal(got, exp)


def test_cumsum():
    """Test `quaxed.lax.cumsum`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.cumsum(x, axis=0)
    exp = lax.cumsum(x, axis=0)

    assert jnp.array_equal(got, exp)


def test_digamma():
    """Test `quaxed.lax.digamma`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float) / 10
    got = qlax.digamma(x)
    exp = lax.digamma(x)

    assert jnp.array_equal(got, exp)


def test_div():
    """Test `quaxed.lax.div`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    y = jnp.array([[5, 6], [7, 8]], dtype=float)
    got = qlax.div(x, y)
    exp = lax.div(x, y)

    assert jnp.array_equal(got, exp)


def test_dot():
    """Test `quaxed.lax.dot`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    y = jnp.array([[5, 6], [7, 8]], dtype=float)
    got = qlax.dot(x, y)
    exp = lax.dot(x, y)

    assert jnp.array_equal(got, exp)


@pytest.mark.skip(reason="TODO: implement.")
def test_dot_general():
    """Test `quaxed.lax.dot_general`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_dynamic_index_in_dim():
    """Test `quaxed.lax.dynamic_index_in_dim`."""


def test_dynamic_slice():
    """Test `quaxed.lax.dynamic_slice`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.dynamic_slice(x, (0, 0), (2, 2))
    exp = lax.dynamic_slice(x, (0, 0), (2, 2))

    assert jnp.array_equal(got, exp)


@pytest.mark.skip(reason="TODO: implement.")
def test_dynamic_slice_in_dim():
    """Test `quaxed.lax.dynamic_slice_in_dim`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_dynamic_update_index_in_dim():
    """Test `quaxed.lax.dynamic_update_index_in_dim`."""


def test_dynamic_update_slice():
    """Test `quaxed.lax.dynamic_update_slice`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    y = jnp.array([[5, 6], [7, 8]], dtype=float)
    got = qlax.dynamic_update_slice(x, y, (0, 0))
    exp = lax.dynamic_update_slice(x, y, (0, 0))

    assert jnp.array_equal(got, exp)


def test_dynamics_update_slice_in_dim():
    """Test `quaxed.lax.dynamic_update_slice_in_dim`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    y = jnp.array([[5, 6], [7, 8]], dtype=float)
    got = qlax.dynamic_update_slice_in_dim(x, y, 0, 0)
    exp = lax.dynamic_update_slice_in_dim(x, y, 0, 0)

    assert jnp.array_equal(got, exp)


def test_eq():
    """Test `quaxed.lax.eq`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.eq(x, x)
    exp = lax.eq(x, x)

    assert jnp.array_equal(got, exp)


def test_erf():
    """Test `quaxed.lax.erf`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float) / 10
    got = qlax.erf(x)
    exp = lax.erf(x)

    assert jnp.array_equal(got, exp)


def test_erfc():
    """Test `quaxed.lax.erfc`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float) / 10
    got = qlax.erfc(x)
    exp = lax.erfc(x)

    assert jnp.array_equal(got, exp)


def test_erf_inv():
    """Test `quaxed.lax.erf_inv`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float) / 10
    got = qlax.erf_inv(x)
    exp = lax.erf_inv(x)

    assert jnp.array_equal(got, exp)


def test_exp():
    """Test `quaxed.lax.exp`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.exp(x)
    exp = lax.exp(x)

    assert jnp.array_equal(got, exp)


def test_expand_dims():
    """Test `quaxed.lax.expand_dims`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.expand_dims(x, (0,))
    exp = lax.expand_dims(x, (0,))

    assert jnp.array_equal(got, exp)


def test_expm1():
    """Test `quaxed.lax.expm1`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.expm1(x)
    exp = lax.expm1(x)

    assert jnp.array_equal(got, exp)


def test_fft():
    """Test `quaxed.lax.fft`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.fft(x, fft_type="fft", fft_lengths=(2, 2))
    exp = lax.fft(x, fft_type="fft", fft_lengths=(2, 2))

    assert jnp.array_equal(got, exp)


def test_floor():
    """Test `quaxed.lax.floor`."""
    x = jnp.array([[1.1, 2.2], [3.3, 4.4]], dtype=float)
    got = qlax.floor(x)
    exp = lax.floor(x)

    assert jnp.array_equal(got, exp)


def test_full():
    """Test `quaxed.lax.full`."""
    got = qlax.full((2, 2), 1.0)
    exp = lax.full((2, 2), 1.0)

    assert jnp.array_equal(got, exp)


def test_full_like():
    """Test `quaxed.lax.full_like`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.full_like(x, 1.0)
    exp = lax.full_like(x, 1.0)

    assert jnp.array_equal(got, exp)


@pytest.mark.skip(reason="TODO: implement.")
def test_gather():
    """Test `quaxed.lax.gather`."""


def test_ge():
    """Test `quaxed.lax.ge`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    y = jnp.array([[5, 2], [7, 2]], dtype=float)
    got = qlax.ge(x, y)
    exp = lax.ge(x, y)

    assert jnp.array_equal(got, exp)


def test_gt():
    """Test `quaxed.lax.gt`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    y = jnp.array([[5, 1], [7, 2]], dtype=float)
    got = qlax.gt(x, y)
    exp = lax.gt(x, y)

    assert jnp.array_equal(got, exp)


def test_igamma():
    """Test `quaxed.lax.igamma`."""
    a = 1.0
    x = jnp.array([[1, 2], [3, 4]], dtype=float) / 10
    got = qlax.igamma(a, x)
    exp = lax.igamma(a, x)

    assert jnp.array_equal(got, exp)


def test_igammac():
    """Test `quaxed.lax.igammac`."""
    a = 1.0
    x = jnp.array([[1, 2], [3, 4]], dtype=float) / 10
    got = qlax.igammac(a, x)
    exp = lax.igammac(a, x)

    assert jnp.array_equal(got, exp)


def test_imag():
    """Test `quaxed.lax.imag`."""
    x = jnp.array([[1 + 1j, 2 + 2j], [3 + 3j, 4 + 4j]], dtype=complex)
    got = qlax.imag(x)
    exp = lax.imag(x)

    assert jnp.array_equal(got, exp)


def test_index_in_dim():
    """Test `quaxed.lax.index_in_dim`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.index_in_dim(x, 0, 0)
    exp = lax.index_in_dim(x, 0, 0)

    assert jnp.array_equal(got, exp)


@pytest.mark.skip(reason="TODO: implement.")
def test_index_take():
    """Test `quaxed.lax.index_take`."""


def test_integer_pow():
    """Test `quaxed.lax.integer_pow`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.integer_pow(x, 2)
    exp = lax.integer_pow(x, 2)

    assert jnp.array_equal(got, exp)


@pytest.mark.skip(reason="TODO: implement.")
def test_iota():
    """Test `quaxed.lax.iota`."""


def test_is_finite():
    """Test `quaxed.lax.is_finite`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.is_finite(x)
    exp = lax.is_finite(x)

    assert jnp.array_equal(got, exp)


def test_le():
    """Test `quaxed.lax.le`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    y = jnp.array([[5, 2], [7, 2]], dtype=float)
    got = qlax.le(x, y)
    exp = lax.le(x, y)

    assert jnp.array_equal(got, exp)


def test_lgamma():
    """Test `quaxed.lax.lgamma`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.lgamma(x)
    exp = lax.lgamma(x)

    assert jnp.array_equal(got, exp)


def test_log():
    """Test `quaxed.lax.log`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.log(x)
    exp = lax.log(x)

    assert jnp.array_equal(got, exp)


def test_log1p():
    """Test `quaxed.lax.log1p`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.log1p(x)
    exp = lax.log1p(x)

    assert jnp.array_equal(got, exp)


def test_logistic():
    """Test `quaxed.lax.logistic`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.logistic(x)
    exp = lax.logistic(x)

    assert jnp.array_equal(got, exp)


def test_lt():
    """Test `quaxed.lax.lt`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    y = jnp.array([[5, 1], [7, 2]], dtype=float)
    got = qlax.lt(x, y)
    exp = lax.lt(x, y)

    assert jnp.array_equal(got, exp)


def test_max():
    """Test `quaxed.lax.max`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    y = jnp.array([[5, 6], [7, 8]], dtype=float)
    got = qlax.max(x, y)
    exp = lax.max(x, y)

    assert jnp.array_equal(got, exp)


def test_min():
    """Test `quaxed.lax.min`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    y = jnp.array([[5, 6], [7, 8]], dtype=float)
    got = qlax.min(x, y)
    exp = lax.min(x, y)

    assert jnp.array_equal(got, exp)


def test_mul():
    """Test `quaxed.lax.mul`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    y = jnp.array([[5, 6], [7, 8]], dtype=float)
    got = qlax.mul(x, y)
    exp = lax.mul(x, y)

    assert jnp.array_equal(got, exp)


def test_ne():
    """Test `quaxed.lax.ne`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    y = jnp.array([[5, 2], [7, 2]], dtype=float)
    got = qlax.ne(x, y)
    exp = lax.ne(x, y)

    assert jnp.array_equal(got, exp)


def test_neg():
    """Test `quaxed.lax.neg`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.neg(x)
    exp = lax.neg(x)

    assert jnp.array_equal(got, exp)


def test_nextafter():
    """Test `quaxed.lax.nextafter`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    y = jnp.array([[5, 6], [7, 8]], dtype=float)
    got = qlax.nextafter(x, y)
    exp = lax.nextafter(x, y)

    assert jnp.array_equal(got, exp)


@pytest.mark.skip(reason="TODO: implement.")
def test_pad():
    """Test `quaxed.lax.pad`."""


def test_polygamma():
    """Test `quaxed.lax.polygamma`."""
    m = 1.0
    x = jnp.array([[1, 2], [3, 4]], dtype=float) / 10
    got = qlax.polygamma(m, x)
    exp = lax.polygamma(m, x)

    assert jnp.array_equal(got, exp)


def test_population_count():
    """Test `quaxed.lax.population_count`."""
    x = jnp.array([[1, 0], [0, 1]], dtype=int)
    got = qlax.population_count(x)
    exp = lax.population_count(x)

    assert jnp.array_equal(got, exp)


def test_pow():
    """Test `quaxed.lax.pow`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    y = jnp.array([[5, 6], [7, 8]], dtype=float)
    got = qlax.pow(x, y)
    exp = lax.pow(x, y)

    assert jnp.array_equal(got, exp)


def test_random_gamma_grad():
    """Test `quaxed.lax.random_gamma_grad`."""
    a = 1.0
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.random_gamma_grad(a, x)
    exp = lax.random_gamma_grad(a, x)

    assert jnp.array_equal(got, exp)


def test_real():
    """Test `quaxed.lax.real`."""
    x = jnp.array([[1 + 1j, 2 + 2j], [3 + 3j, 4 + 4j]], dtype=complex)
    got = qlax.real(x)
    exp = lax.real(x)

    assert jnp.array_equal(got, exp)


def test_reciprocal():
    """Test `quaxed.lax.reciprocal`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.reciprocal(x)
    exp = lax.reciprocal(x)

    assert jnp.array_equal(got, exp)


@pytest.mark.skip(reason="TODO: implement.")
def test_reduce():
    """Test `quaxed.lax.reduce`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_reduce_precision():
    """Test `quaxed.lax.reduce_precision`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_reduce_window():
    """Test `quaxed.lax.reduce_window`."""


def test_rem():
    """Test `quaxed.lax.rem`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    y = jnp.array([[5, 6], [7, 8]], dtype=float)
    got = qlax.rem(x, y)
    exp = lax.rem(x, y)

    assert jnp.array_equal(got, exp)


def test_reshape():
    """Test `quaxed.lax.reshape`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.reshape(x, (1, 4))
    exp = lax.reshape(x, (1, 4))

    assert jnp.array_equal(got, exp)


def test_rev():
    """Test `quaxed.lax.rev`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.rev(x, dimensions=(0,))
    exp = lax.rev(x, dimensions=(0,))

    assert jnp.array_equal(got, exp)


@pytest.mark.skip(reason="TODO: implement.")
def test_rng_bit_generator():
    """Test `quaxed.lax.rng_bit_generator`."""


def test_rng_uniform():
    """Test `quaxed.lax.rng_uniform`."""
    a, b = 0, 1
    got = qlax.rng_uniform(a, b, (2, 3))
    exp = lax.rng_uniform(a, b, (2, 3))

    assert jnp.array_equal(got, exp)


def test_round():
    """Test `quaxed.lax.round`."""
    x = jnp.array([[1.1, 2.2], [3.3, 4.4]], dtype=float)
    got = qlax.round(x)
    exp = lax.round(x)

    assert jnp.array_equal(got, exp)


def test_rsqrt():
    """Test `quaxed.lax.rsqrt`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.rsqrt(x)
    exp = lax.rsqrt(x)

    assert jnp.array_equal(got, exp)


@pytest.mark.skip(reason="TODO: implement.")
def test_scatter():
    """Test `quaxed.lax.scatter`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_scatter_apply():
    """Test `quaxed.lax.scatter_apply`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_scatter_max():
    """Test `quaxed.lax.scatter_max`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_scatter_min():
    """Test `quaxed.lax.scatter_min`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_scatter_mul():
    """Test `quaxed.lax.scatter_mul`."""


def test_shift_left():
    """Test `quaxed.lax.shift_left`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=int)
    got = qlax.shift_left(x, 1)
    exp = lax.shift_left(x, 1)

    assert jnp.array_equal(got, exp)


def test_shift_right_arithmetic():
    """Test `quaxed.lax.shift_right_arithmetic`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=int)
    got = qlax.shift_right_arithmetic(x, 1)
    exp = lax.shift_right_arithmetic(x, 1)

    assert jnp.array_equal(got, exp)


def test_shift_right_logical():
    """Test `quaxed.lax.shift_right_logical`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=int)
    got = qlax.shift_right_logical(x, 1)
    exp = lax.shift_right_logical(x, 1)

    assert jnp.array_equal(got, exp)


def test_sign():
    """Test `quaxed.lax.sign`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.sign(x)
    exp = lax.sign(x)

    assert jnp.array_equal(got, exp)


def test_sin():
    """Test `quaxed.lax.sin`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.sin(x)
    exp = lax.sin(x)

    assert jnp.array_equal(got, exp)


def test_sinh():
    """Test `quaxed.lax.sinh`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.sinh(x)
    exp = lax.sinh(x)

    assert jnp.array_equal(got, exp)


def test_slice():
    """Test `quaxed.lax.slice`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.slice(x, (0, 0), (2, 2))
    exp = lax.slice(x, (0, 0), (2, 2))

    assert jnp.array_equal(got, exp)


def test_slice_in_dim():
    """Test `quaxed.lax.slice_in_dim`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.slice_in_dim(x, 0, 0, 2)
    exp = lax.slice_in_dim(x, 0, 0, 2)

    assert jnp.array_equal(got, exp)


def test_sort():
    """Test `quaxed.lax.sort`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.sort(x)
    exp = lax.sort(x)

    assert jnp.array_equal(got, exp)


@pytest.mark.skip(reason="TODO: implement.")
def test_sort_key_val():
    """Test `quaxed.lax.sort_key_val`."""


def test_sqrt():
    """Test `quaxed.lax.sqrt`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.sqrt(x)
    exp = lax.sqrt(x)

    assert jnp.array_equal(got, exp)


def test_square():
    """Test `quaxed.lax.square`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.square(x)
    exp = lax.square(x)

    assert jnp.array_equal(got, exp)


def test_sub():
    """Test `quaxed.lax.sub`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    y = jnp.array([[5, 6], [7, 8]], dtype=float)
    got = qlax.sub(x, y)
    exp = lax.sub(x, y)

    assert jnp.array_equal(got, exp)


def test_tan():
    """Test `quaxed.lax.tan`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.tan(x)
    exp = lax.tan(x)

    assert jnp.array_equal(got, exp)


def test_tanh():
    """Test `quaxed.lax.tanh`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.tanh(x)
    exp = lax.tanh(x)

    assert jnp.array_equal(got, exp)


def test_top_k():
    """Test `quaxed.lax.top_k`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.top_k(x, 1)
    exp = lax.top_k(x, 1)

    assert jnp.array_equal(got, exp)


def test_transpose():
    """Test `quaxed.lax.transpose`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.transpose(x, (1, 0))
    exp = lax.transpose(x, (1, 0))

    assert jnp.array_equal(got, exp)


def test_zeros_like_array():
    """Test `quaxed.lax.zeros_like_array`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.zeros_like_array(x)
    exp = lax.zeros_like_array(x)

    assert jnp.array_equal(got, exp)


def test_zeta():
    """Test `quaxed.lax.zeta`."""
    x = jnp.array([[5, 2], [3, 4]], dtype=float)
    q = 2.0
    got = qlax.zeta(x, q)
    exp = lax.zeta(x, q)

    assert jnp.array_equal(got, exp)


# ==============================================================================
# Control flow operators


@pytest.mark.skip(reason="TODO: implement.")
def test_associative_scan():
    """Test `quaxed.lax.associative_scan`."""


def test_cond():
    """Test `quaxed.lax.cond`."""
    pred = True
    true_fun = lambda: jnp.array([[1, 2], [3, 4]], dtype=float)  # noqa: E731  # pylint: disable=C3001
    false_fun = lambda: jnp.array([[5, 6], [7, 8]], dtype=float)  # noqa: E731 # pylint: disable=C3001
    got = qlax.cond(pred, true_fun, false_fun)
    exp = lax.cond(pred, true_fun, false_fun)

    assert jnp.array_equal(got, exp)


@pytest.mark.skip(reason="TODO: implement.")
def test_fori_loop():
    """Test `quaxed.lax.fori_loop`."""


def test_map():
    """Test `quaxed.lax.map`."""
    fun = lambda x: x + 1  # noqa: E731  # pylint: disable=C3001
    xs = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.map(fun, xs)
    exp = lax.map(fun, xs)

    assert jnp.array_equal(got, exp)


@pytest.mark.skip(reason="TODO: implement.")
def test_scan():
    """Test `quaxed.lax.scan`."""


def test_select():
    """Test `quaxed.lax.select`."""
    pred = jnp.array([[True, False], [True, False]], dtype=bool)
    on_true = jnp.array([[1, 2], [3, 4]], dtype=float)
    on_false = jnp.array([[5, 6], [7, 8]], dtype=float)
    got = qlax.select(pred, on_true, on_false)
    exp = lax.select(pred, on_true, on_false)

    assert jnp.array_equal(got, exp)


@pytest.mark.skip(reason="TODO: implement.")
def test_select_n():
    """Test `quaxed.lax.select_n`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_switch():
    """Test `quaxed.lax.switch`."""


def test_while_loop():
    """Test `quaxed.lax.while_loop`."""
    cond_fun = lambda x: jnp.all(x < 10)  # noqa: E731  # pylint: disable=C3001
    body_fun = lambda x: x + 1  # noqa: E731  # pylint: disable=C3001
    init_val = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.while_loop(cond_fun, body_fun, init_val)
    exp = lax.while_loop(cond_fun, body_fun, init_val)

    assert jnp.array_equal(got, exp)


# ==============================================================================
# Custom gradient operators


def test_stop_gradient():
    """Test `quaxed.lax.stop_gradient`."""
    x = jnp.array([[1, 2], [3, 4]], dtype=float)
    got = qlax.stop_gradient(x)
    exp = lax.stop_gradient(x)

    assert jnp.array_equal(got, exp)


@pytest.mark.skip(reason="TODO: implement.")
def test_custom_linear_solve():
    """Test `quaxed.lax.custom_linear_solve`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_custom_root():
    """Test `quaxed.lax.custom_root`."""


# ==============================================================================
# Parallel operators


@pytest.mark.skip(reason="TODO: implement.")
def test_all_gather():
    """Test `quaxed.lax.all_gather`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_all_to_all():
    """Test `quaxed.lax.all_to_all`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_psum():
    """Test `quaxed.lax.psum`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_psum_scatter():
    """Test `quaxed.lax.psum_scatter`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_pmax():
    """Test `quaxed.lax.pmax`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_pmin():
    """Test `quaxed.lax.pmin`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_pmin():
    """Test `quaxed.lax.pmin`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_pmean():
    """Test `quaxed.lax.pmean`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_ppermute():
    """Test `quaxed.lax.ppermute`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_pshuffle():
    """Test `quaxed.lax.pshuffle`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_pswapaxes():
    """Test `quaxed.lax.pswapaxes`."""


@pytest.mark.skip(reason="TODO: implement.")
def test_axis_index():
    """Test `quaxed.lax.axis_index`."""


# ==============================================================================
# Sharding-related operators


@pytest.mark.skip(reason="TODO: implement.")
def test_with_sharding_constraint():
    """Test `quaxed.lax.with_sharding_constraint`."""


# ==============================================================================
# Linear algebra operators


def test_cholesky():
    """Test `quaxed.lax.cholesky`."""
    x = jnp.array([[1, 2], [2, 5]], dtype=float)
    got = qlax.linalg.cholesky(x)
    exp = lax.linalg.cholesky(x)

    assert jnp.array_equal(got, exp)


def test_eig():
    """Test `quaxed.lax.eig`."""
    x = jnp.array([[1, 2], [2, 5]], dtype=float)
    got_w, got_vl, got_vr = lax.linalg.eig(x)
    exp_w, exp_vl, exp_vr = qlax.linalg.eig(x)

    assert jnp.array_equal(got_w, exp_w)
    assert jnp.array_equal(got_vl, exp_vl)
    assert jnp.array_equal(got_vr, exp_vr)


def test_eigh():
    """Test `quaxed.lax.eigh`."""
    x = jnp.array([[1, 2], [2, 5]], dtype=float)
    got_w, got_v = lax.linalg.eigh(x)
    exp_w, exp_v = qlax.linalg.eigh(x)

    assert jnp.array_equal(got_w, exp_w)
    assert jnp.array_equal(got_v, exp_v)


def test_hessenberg():
    """Test `quaxed.lax.hessenberg`."""
    x = jnp.array([[1, 2], [2, 5]], dtype=float)
    got_A, got_taus = lax.linalg.hessenberg(x)  # noqa: N806
    exp_A, exp_taus = qlax.linalg.hessenberg(x)  # noqa: N806

    assert jnp.array_equal(got_A, exp_A)
    assert jnp.array_equal(got_taus, exp_taus)


def test_lu():
    """Test `quaxed.lax.lu`."""
    x = jnp.array([[1, 2], [2, 5]], dtype=float)
    got_lu, got_pivots, got_permutation = lax.linalg.lu(x)
    exp_lu, exp_pivots, exp_permutation = qlax.linalg.lu(x)

    assert jnp.array_equal(got_lu, exp_lu)
    assert jnp.array_equal(got_pivots, exp_pivots)
    assert jnp.array_equal(got_permutation, exp_permutation)


def test_householder_product():
    """Test `quaxed.lax.householder_product`."""
    x = jnp.array([[1, 2], [2, 5]], dtype=float)
    A, taus = lax.linalg.hessenberg(x)  # noqa: N806
    got = qlax.linalg.householder_product(A, taus)
    exp = lax.linalg.householder_product(A, taus)

    assert jnp.array_equal(got, exp)


def test_qdwh():
    """Test `quaxed.lax.qdwh`."""
    x = jnp.array([[1, 2], [2, 5]], dtype=float)
    got_u, got_h, got_num_iters, got_is_converged = lax.linalg.qdwh(x)
    exp_u, exp_h, exp_num_iters, exp_is_converged = qlax.linalg.qdwh(x)

    assert jnp.array_equal(got_u, exp_u)
    assert jnp.array_equal(got_h, exp_h)
    assert jnp.array_equal(got_num_iters, exp_num_iters)
    assert jnp.array_equal(got_is_converged, exp_is_converged)


def test_qr():
    """Test `quaxed.lax.linalg.qr`."""
    x = jnp.array([[1, 2], [2, 5]], dtype=float)
    got = qlax.linalg.qr(x)
    exp = lax.linalg.qr(x)

    assert jnp.array_equal(got, exp)


def test_schur():
    """Test `quaxed.lax.linalg.schur`."""
    x = jnp.array([[1, 2], [2, 5]], dtype=float)
    got = qlax.linalg.schur(x)
    exp = lax.linalg.schur(x)

    assert jnp.array_equal(got, exp)


def test_svd():
    """Test `quaxed.lax.svd`."""
    x = jnp.array([[1, 2], [2, 5]], dtype=float)
    got_u, got_s, got_v = lax.linalg.svd(x)
    exp_u, exp_s, exp_v = qlax.linalg.svd(x)

    assert jnp.array_equal(got_u, exp_u)
    assert jnp.array_equal(got_s, exp_s)
    assert jnp.array_equal(got_v, exp_v)


def test_triangular_solve():
    """Test `quaxed.lax.triangular_solve`."""
    x = jnp.array([[1, 2], [2, 5]], dtype=float)
    y = jnp.array([[1, 2], [2, 5]], dtype=float)
    got = qlax.linalg.triangular_solve(x, y)
    exp = lax.linalg.triangular_solve(x, y)

    assert jnp.array_equal(got, exp)


def test_tridiagonal():
    """Test `quaxed.lax.tridiagonal`."""
    x = jnp.array([[1, 2], [2, 5]], dtype=float)
    got_a, got_d, got_e, got_taus = qlax.linalg.tridiagonal(x)
    exp_a, exp_d, exp_e, exp_taus = lax.linalg.tridiagonal(x)

    assert jnp.array_equal(got_a, exp_a)
    assert jnp.array_equal(got_d, exp_d)
    assert jnp.array_equal(got_e, exp_e)
    assert jnp.array_equal(got_taus, exp_taus)


@pytest.mark.skip(reason="TODO: implement.")
def test_tridiagonal_solve():
    """Test `quaxed.lax.tridiagonal_solve`."""



from jax.scipy.special import bernoulli as bernoulli
from jax.scipy.special import bessel_jn as bessel_jn
from jax.scipy.special import beta as beta
from jax.scipy.special import betainc as betainc
from jax.scipy.special import betaln as betaln
from jax.scipy.special import digamma as digamma
from jax.scipy.special import entr as entr
from jax.scipy.special import erf as erf
from jax.scipy.special import erfc as erfc
from jax.scipy.special import erfinv as erfinv
from jax.scipy.special import exp1 as exp1
from jax.scipy.special import expi as expi
from jax.scipy.special import expit as expit
from jax.scipy.special import expn as expn
from jax.scipy.special import factorial as factorial
from jax.scipy.special import gamma as gamma
from jax.scipy.special import gammainc as gammainc
from jax.scipy.special import gammaincc as gammaincc
from jax.scipy.special import gammaln as gammaln
from jax.scipy.special import hyp1f1 as hyp1f1
from jax.scipy.special import i0 as i0
from jax.scipy.special import i0e as i0e
from jax.scipy.special import i1 as i1
from jax.scipy.special import i1e as i1e
from jax.scipy.special import kl_div as kl_div
from jax.scipy.special import log_ndtr as log_ndtr
from jax.scipy.special import logit as logit
from jax.scipy.special import logsumexp as logsumexp
from jax.scipy.special import lpmn as lpmn
from jax.scipy.special import lpmn_values as lpmn_values
from jax.scipy.special import multigammaln as multigammaln
from jax.scipy.special import ndtr as ndtr
from jax.scipy.special import ndtri as ndtri
from jax.scipy.special import poch as poch
from jax.scipy.special import polygamma as polygamma
from jax.scipy.special import rel_entr as rel_entr
from jax.scipy.special import spence as spence
from jax.scipy.special import sph_harm as sph_harm
from jax.scipy.special import xlog1py as xlog1py
from jax.scipy.special import xlogy as xlogy
from jax.scipy.special import zeta as zeta



"""Doctest configuration."""

import platform
from doctest import ELLIPSIS, NORMALIZE_WHITESPACE

from sybil import Sybil
from sybil.parsers.rest import DocTestParser, PythonCodeBlockParser, SkipParser

# TODO: stop skipping doctests on Windows when there is uniform support for
#       numpy 2.0+ scalar repr. On windows it is printed as 1.0 instead of
#       `np.float64(1.0)`.
parsers = (
    [DocTestParser(optionflags=ELLIPSIS | NORMALIZE_WHITESPACE)]
    if platform.system() != "Windows"
    else []
) + [
    PythonCodeBlockParser(),
    SkipParser(),
]

pytest_collect_file = Sybil(
    parsers=parsers,
    patterns=["*.rst", "*.py"],
).pytest()



"""Copyright (c) 2023 Nathaniel Starkman. All rights reserved.

quaxed: Pre-quaxed libraries for multiple dispatch over abstract array types in JAX
"""
# pylint: disable=C0415,W0621

from __future__ import annotations

from typing import TYPE_CHECKING

from . import _jax, lax, numpy, scipy
from ._jax import *
from ._setup import JAX_VERSION
from ._version import version as __version__  # noqa: F401

__all__ = ["lax", "numpy", "scipy"]
__all__ += _jax.__all__

if JAX_VERSION < (0, 4, 32):
    from . import array_api

    __all__ += ["array_api"]


if TYPE_CHECKING:
    from typing import Any


def __getattr__(name: str) -> Any:  # TODO: fuller annotation
    """Forward all other attribute accesses to Quaxified JAX."""
    import sys

    import jax
    from quax import quaxify

    # TODO: detect if the attribute is a function or a module.
    # If it is a function, quaxify it. If it is a module, return a proxy object
    # that quaxifies all of its attributes.
    out = quaxify(getattr(jax, name))

    # Cache the function in this module
    setattr(sys.modules[__name__], name, out)

    return out


# Clean up the namespace
del TYPE_CHECKING



"""Sorting functions."""

__all__ = ["argsort", "sort"]


from jax.experimental import array_api
from jaxtyping import ArrayLike
from quax import Value

from quaxed._utils import quaxify


@quaxify
def argsort(
    x: ArrayLike,
    /,
    *,
    axis: int = -1,
    descending: bool = False,
    stable: bool = True,
) -> Value:
    return array_api.argsort(x, axis=axis, descending=descending, stable=stable)


@quaxify
def sort(
    x: ArrayLike,
    /,
    *,
    axis: int = -1,
    descending: bool = False,
    stable: bool = True,
) -> Value:
    return array_api.sort(x, axis=axis, descending=descending, stable=stable)



__all__ = ["unique_all", "unique_counts", "unique_inverse", "unique_values"]


from jax.experimental import array_api
from jaxtyping import ArrayLike
from quax import Value

from quaxed._utils import quaxify


@quaxify
def unique_all(x: ArrayLike, /) -> tuple[Value, Value, Value, Value]:
    return array_api.unique_all(x)


@quaxify
def unique_counts(x: ArrayLike, /) -> tuple[Value, Value]:
    return array_api.unique_counts(x)


@quaxify
def unique_inverse(x: ArrayLike, /) -> tuple[Value, Value]:
    return array_api.unique_inverse(x)


@quaxify
def unique_values(x: ArrayLike, /) -> Value:
    return array_api.unique_values(x)



"""Test with JAX inputs."""

import jax
import jax.numpy as jnp
import pytest

import quaxed


def test_dir():
    """Test the `__dir__` method."""
    assert quaxed.scipy.linalg.__dir__() == quaxed.scipy.linalg.__all__


def test_not_in_scipy_special():
    """Test error message for non-members."""
    with pytest.raises(AttributeError, match="Cannot get"):
        _ = quaxed.scipy.special.not_a_member


def test_not_in_scipy_linalg():
    """Test error message for non-members."""
    with pytest.raises(AttributeError, match="Cannot get"):
        _ = quaxed.scipy.linalg.not_a_member


def test_svd():
    """Test `quaxed.scipy.linalg.svd`."""
    assert hasattr(jax.scipy.linalg, "svd")
    assert hasattr(quaxed.scipy.linalg, "svd")

    x = jnp.array([[1, 2], [3, 4]])
    got = quaxed.scipy.linalg.svd(x, compute_uv=False)
    expected = jax.scipy.linalg.svd(x, compute_uv=False)

    assert jnp.array_equal(got, expected)



"""Quaxed :mod:`jax`."""

__all__ = [
    "device_put",
    "grad",
    "hessian",
    "jacfwd",
    "jacrev",
    "value_and_grad",
]

from collections.abc import Callable, Hashable
from typing import Any, TypeAlias

import jax
from quax import quaxify

AxisName: TypeAlias = Hashable


# =============================================================================

device_put = quaxify(jax.device_put)


# -----------------------------------------------------------------------------


def grad(
    fun: Callable[..., Any],
    *,
    filter_spec: Any = True,
    **kwargs: Any,
) -> Callable[..., Any]:
    """Quaxed version of :func:`jax.grad`.

    This adds the additional parameter ``filter_spec``, which is passed to
    :func:`quax.quaxify`, to support passing integers when ``allow_int`` is
    `True`.

    Examples
    --------
    >>> import jax.numpy as jnp
    >>> from quaxed import grad

    >>> grad(jnp.square)(jnp.array(3.))
    Array(6., dtype=float32, weak_type=True)

    """
    return quaxify(jax.grad(fun, **kwargs), filter_spec=filter_spec)


# -----------------------------------------------------------------------------


def hessian(
    fun: Callable[..., Any],
    *,
    filter_spec: Any = True,
    **kwargs: Any,
) -> Callable[..., Any]:
    """Quaxed version of :func:`jax.hessian`.

    This adds the additional parameter ``filter_spec``, which is passed to
    :func:`quax.quaxify`, to support passing integers when ``allow_int`` is
    `True`.

    Examples
    --------
    >>> import jax.numpy as jnp
    >>> from quaxed import hessian

    >>> hessian(jnp.square)(jnp.array(3.))
    Array(2., dtype=float32, weak_type=True)

    """
    return quaxify(jax.hessian(fun, **kwargs), filter_spec=filter_spec)


# -----------------------------------------------------------------------------


def jacfwd(
    fun: Callable[..., Any],
    *,
    filter_spec: Any = True,
    **kwargs: Any,
) -> Callable[..., Any]:
    """Quaxed version of :func:`jax.jacfwd`.

    This adds the additional parameter ``filter_spec``, which is passed to
    :func:`quax.quaxify`, to support passing integers when ``allow_int`` is
    `True`.

    Examples
    --------
    >>> import jax.numpy as jnp
    >>> from quaxed import jacfwd

    >>> jacfwd(jnp.square)(jnp.array(3.))
    Array(6., dtype=float32, weak_type=True)

    """
    return quaxify(jax.jacfwd(fun, **kwargs), filter_spec=filter_spec)


def jacrev(
    fun: Callable[..., Any],
    *,
    filter_spec: Any = True,
    **kwargs: Any,
) -> Callable[..., Any]:
    """Quaxed version of :func:`jax.jacfwd`.

    This adds the additional parameter ``filter_spec``, which is passed to
    :func:`quax.quaxify`, to support passing integers when ``allow_int`` is
    `True`.

    Examples
    --------
    >>> import jax.numpy as jnp
    >>> from quaxed import jacrev

    >>> jacrev(jnp.square)(jnp.array(3.))
    Array(6., dtype=float32, weak_type=True)

    """
    return quaxify(jax.jacrev(fun, **kwargs), filter_spec=filter_spec)


# -----------------------------------------------------------------------------


def value_and_grad(
    fun: Callable[..., Any],
    *,
    filter_spec: Any = True,
    **kwargs: Any,
) -> Callable[..., Any]:
    """Quaxed version of :func:`jax.value_and_grad`.

    This adds the additional parameter ``filter_spec``, which is passed to
    :func:`quax.quaxify`, to support passing integers when ``allow_int`` is
    `True`.

    Examples
    --------
    >>> import jax.numpy as jnp
    >>> from quaxed import value_and_grad

    >>> value_and_grad(jnp.square)(jnp.array(3.))
    (Array(9., dtype=float32, weak_type=True),
     Array(6., dtype=float32, weak_type=True))

    """
    return quaxify(jax.value_and_grad(fun, **kwargs), filter_spec=filter_spec)



"""Quaxed :mod:`jax.numpy.linalg`."""

__all__ = [  # noqa: F822
    "cholesky",
    "cross",
    "det",
    "diagonal",
    "eig",
    "eigh",
    "eigvalsh",
    "inv",
    "matmul",
    "matrix_norm",
    "matrix_power",
    "matrix_rank",
    "matrix_transpose",
    "outer",
    "pinv",
    "qr",
    "slogdet",
    "solve",
    "svd",
    "svdvals",
    "tensordot",
    "trace",
    "vecdot",
    "vector_norm",
    "norm",
]

import sys
from collections.abc import Callable
from typing import Any

import jax.numpy as jnp
from quax import quaxify


def __dir__() -> list[str]:
    return sorted(__all__)


def __getattr__(name: str) -> Callable[..., Any]:  # TODO: better type hint
    """Get the object from the `jax.numpy` module."""
    if name not in __all__:
        msg = f"module {__name__!r} has no attribute {name!r}"
        raise AttributeError(msg)

    # Get the object
    jnp_obj = getattr(jnp.linalg, name)

    # Quaxify?
    out = quaxify(jnp_obj)

    # Cache the function in this module
    setattr(sys.modules[__name__], name, out)

    return out



# ----- Operators -----
# isort: split
from jax.lax.linalg import abs as abs
from jax.lax.linalg import acos as acos
from jax.lax.linalg import acosh as acosh
from jax.lax.linalg import add as add
from jax.lax.linalg import approx_max_k as approx_max_k
from jax.lax.linalg import approx_min_k as approx_min_k
from jax.lax.linalg import argmax as argmax
from jax.lax.linalg import argmin as argmin
from jax.lax.linalg import asin as asin
from jax.lax.linalg import asinh as asinh
from jax.lax.linalg import atan as atan
from jax.lax.linalg import atan2 as atan2
from jax.lax.linalg import atanh as atanh
from jax.lax.linalg import batch_matmul as batch_matmul
from jax.lax.linalg import bessel_i0e as bessel_i0e
from jax.lax.linalg import bessel_i1e as bessel_i1e
from jax.lax.linalg import betainc as betainc
from jax.lax.linalg import bitcast_convert_type as bitcast_convert_type
from jax.lax.linalg import bitwise_and as bitwise_and
from jax.lax.linalg import bitwise_not as bitwise_not
from jax.lax.linalg import bitwise_or as bitwise_or
from jax.lax.linalg import bitwise_xor as bitwise_xor
from jax.lax.linalg import broadcast as broadcast
from jax.lax.linalg import broadcast_in_dim as broadcast_in_dim
from jax.lax.linalg import broadcast_shapes as broadcast_shapes
from jax.lax.linalg import broadcast_to_rank as broadcast_to_rank
from jax.lax.linalg import broadcasted_iota as broadcasted_iota
from jax.lax.linalg import cbrt as cbrt
from jax.lax.linalg import ceil as ceil
from jax.lax.linalg import clamp as clamp
from jax.lax.linalg import clz as clz
from jax.lax.linalg import collapse as collapse
from jax.lax.linalg import complex as complex
from jax.lax.linalg import concatenate as concatenate
from jax.lax.linalg import conj as conj
from jax.lax.linalg import conv as conv
from jax.lax.linalg import conv_dimension_numbers as conv_dimension_numbers
from jax.lax.linalg import conv_general_dilated as conv_general_dilated
from jax.lax.linalg import conv_general_dilated_local as conv_general_dilated_local
from jax.lax.linalg import conv_general_dilated_patches as conv_general_dilated_patches
from jax.lax.linalg import conv_transpose as conv_transpose
from jax.lax.linalg import conv_with_general_padding as conv_with_general_padding
from jax.lax.linalg import convert_element_type as convert_element_type
from jax.lax.linalg import cos as cos
from jax.lax.linalg import cosh as cosh
from jax.lax.linalg import cumlogsumexp as cumlogsumexp
from jax.lax.linalg import cummax as cummax
from jax.lax.linalg import cummin as cummin
from jax.lax.linalg import cumprod as cumprod
from jax.lax.linalg import cumsum as cumsum
from jax.lax.linalg import digamma as digamma
from jax.lax.linalg import div as div
from jax.lax.linalg import dot as dot
from jax.lax.linalg import dot_general as dot_general
from jax.lax.linalg import dynamic_index_in_dim as dynamic_index_in_dim
from jax.lax.linalg import dynamic_slice as dynamic_slice
from jax.lax.linalg import dynamic_slice_in_dim as dynamic_slice_in_dim
from jax.lax.linalg import dynamic_update_index_in_dim as dynamic_update_index_in_dim
from jax.lax.linalg import dynamic_update_slice as dynamic_update_slice
from jax.lax.linalg import dynamic_update_slice_in_dim as dynamic_update_slice_in_dim
from jax.lax.linalg import eq as eq
from jax.lax.linalg import erf as erf
from jax.lax.linalg import erf_inv as erf_inv
from jax.lax.linalg import erfc as erfc
from jax.lax.linalg import exp as exp
from jax.lax.linalg import expand_dims as expand_dims
from jax.lax.linalg import expm1 as expm1
from jax.lax.linalg import fft as fft
from jax.lax.linalg import floor as floor
from jax.lax.linalg import full as full
from jax.lax.linalg import full_like as full_like
from jax.lax.linalg import gather as gather
from jax.lax.linalg import ge as ge
from jax.lax.linalg import gt as gt
from jax.lax.linalg import igamma as igamma
from jax.lax.linalg import igammac as igammac
from jax.lax.linalg import imag as imag
from jax.lax.linalg import index_in_dim as index_in_dim
from jax.lax.linalg import index_take as index_take
from jax.lax.linalg import integer_pow as integer_pow
from jax.lax.linalg import iota as iota
from jax.lax.linalg import is_finite as is_finite
from jax.lax.linalg import le as le
from jax.lax.linalg import lgamma as lgamma
from jax.lax.linalg import log as log
from jax.lax.linalg import log1p as log1p
from jax.lax.linalg import logistic as logistic
from jax.lax.linalg import lt as lt
from jax.lax.linalg import max as max
from jax.lax.linalg import min as min
from jax.lax.linalg import mul as mul
from jax.lax.linalg import neg as neg
from jax.lax.linalg import nextafter as nextafter
from jax.lax.linalg import pad as pad
from jax.lax.linalg import polygamma as polygamma
from jax.lax.linalg import population_count as population_count
from jax.lax.linalg import pow as pow
from jax.lax.linalg import random_gamma_grad as random_gamma_grad
from jax.lax.linalg import real as real
from jax.lax.linalg import reciprocal as reciprocal
from jax.lax.linalg import reduce as reduce
from jax.lax.linalg import reduce_precision as reduce_precision
from jax.lax.linalg import reduce_window as reduce_window
from jax.lax.linalg import rem as rem
from jax.lax.linalg import reshape as reshape
from jax.lax.linalg import rev as rev
from jax.lax.linalg import rng_bit_generator as rng_bit_generator
from jax.lax.linalg import rng_uniform as rng_uniform
from jax.lax.linalg import round as round
from jax.lax.linalg import rsqrt as rsqrt
from jax.lax.linalg import scatter as scatter
from jax.lax.linalg import scatter_add as scatter_add
from jax.lax.linalg import scatter_apply as scatter_apply
from jax.lax.linalg import scatter_max as scatter_max
from jax.lax.linalg import scatter_min as scatter_min
from jax.lax.linalg import scatter_mul as scatter_mul
from jax.lax.linalg import shift_left as shift_left
from jax.lax.linalg import shift_right_arithmetic as shift_right_arithmetic
from jax.lax.linalg import shift_right_logical as shift_right_logical
from jax.lax.linalg import sign as sign
from jax.lax.linalg import sin as sin
from jax.lax.linalg import sinh as sinh
from jax.lax.linalg import slice as slice
from jax.lax.linalg import slice_in_dim as slice_in_dim
from jax.lax.linalg import sort as sort
from jax.lax.linalg import sort_key_val as sort_key_val
from jax.lax.linalg import sqrt as sqrt
from jax.lax.linalg import square as square
from jax.lax.linalg import squeeze as squeeze
from jax.lax.linalg import sub as sub
from jax.lax.linalg import tan as tan
from jax.lax.linalg import tanh as tanh
from jax.lax.linalg import top_k as top_k
from jax.lax.linalg import transpose as transpose
from jax.lax.linalg import zeros_like_array as zeros_like_array
from jax.lax.linalg import zeta as zeta

# ----- Control Flow Operators -----
# isort: split
from jax.lax.linalg import associative_scan as associative_scan
from jax.lax.linalg import cond as cond
from jax.lax.linalg import fori_loop as fori_loop
from jax.lax.linalg import map as map
from jax.lax.linalg import scan as scan
from jax.lax.linalg import select as select
from jax.lax.linalg import select_n as select_n
from jax.lax.linalg import switch as switch
from jax.lax.linalg import while_loop as while_loop

# ----- Custom Gradient Operators -----
# isort: split
from jax.lax.linalg import custom_linear_solve as custom_linear_solve
from jax.lax.linalg import custom_root as custom_root
from jax.lax.linalg import stop_gradient as stop_gradient

# ----- Parallel Operators -----
# isort: split
from jax.lax.linalg import all_gather as all_gather
from jax.lax.linalg import all_to_all as all_to_all
from jax.lax.linalg import axis_index as axis_index
from jax.lax.linalg import pmax as pmax
from jax.lax.linalg import pmean as pmean
from jax.lax.linalg import pmin as pmin
from jax.lax.linalg import ppermute as ppermute
from jax.lax.linalg import pshuffle as pshuffle
from jax.lax.linalg import psum as psum
from jax.lax.linalg import psum_scatter as psum_scatter
from jax.lax.linalg import pswapaxes as pswapaxes

# ----- Sharding-related Operators -----
# isort: split
from jax.lax.linalg import with_sharding_constraint as with_sharding_constraint

# ----- Linear Algebra Operators -----
# isort: split
from jax.lax.linalg import linalg as linalg



"""Copyright (c) 2023 Nathaniel Starkman. All rights reserved."""

__all__: list[str] = []

from typing import Any, Protocol, runtime_checkable

import jax.numpy as jnp
import plum
from jaxtyping import ArrayLike

# Simplify the display of ArrayLike
plum.activate_union_aliases()
plum.set_union_alias(ArrayLike, "ArrayLike")


@runtime_checkable
class DType(Protocol):
    """The dtype of an array."""

    dtype: jnp.dtype[Any]



"""Array API creation functions."""

__all__ = [
    "arange",
    "asarray",
    "empty",
    "empty_like",
    "eye",
    "from_dlpack",
    "full",
    "full_like",
    "linspace",
    "meshgrid",
    "ones",
    "ones_like",
    "tril",
    "triu",
    "zeros",
    "zeros_like",
]


from typing import TypeVar

import jax
import jax.numpy as jnp
from jax import Device
from jax.experimental import array_api
from jaxtyping import ArrayLike
from quax import Value

from quaxed._types import DType
from quaxed._utils import quaxify

from ._dispatch import dispatcher

T = TypeVar("T")

# =============================================================================


@dispatcher
def arange(
    start: ArrayLike,
    stop: ArrayLike | None,
    step: ArrayLike | None,
    /,
    *,
    dtype: DType | None = None,
    device: Device | None = None,
) -> ArrayLike:
    return array_api.arange(start, stop, step, dtype=dtype, device=device)


@dispatcher  # type: ignore[no-redef]
def arange(
    start: ArrayLike,
    stop: ArrayLike | None,
    *,
    step: ArrayLike | None = None,
    dtype: DType | None = None,
    device: Device | None = None,
) -> ArrayLike:
    # re-dispatch on `start`, `stop`, and `step`
    return arange(start, stop, step, dtype=dtype, device=device)


@dispatcher  # type: ignore[no-redef]
def arange(
    start: ArrayLike,
    *,
    stop: ArrayLike | None = None,
    step: ArrayLike | None = None,
    dtype: DType | None = None,
    device: Device | None = None,
) -> ArrayLike:
    # re- dispatch on `start`, `stop`, and `step`
    return arange(start, stop, step, dtype=dtype, device=device)


@dispatcher  # type: ignore[no-redef]
def arange(
    *,
    start: ArrayLike,
    stop: ArrayLike | None = None,
    step: ArrayLike | None = None,
    dtype: DType | None = None,
    device: Device | None = None,
) -> ArrayLike | Value:
    # re-dispatch on `start`, `stop`, and `step`
    return arange(start, stop, step, dtype=dtype, device=device)


# =============================================================================


@quaxify
def asarray(
    obj: ArrayLike,
    /,
    *,
    dtype: DType | None = None,
    device: Device | None = None,
    copy: bool | None = None,
) -> Value:
    return array_api.asarray(obj, dtype=dtype, device=device, copy=copy)


# =============================================================================


def empty(
    shape: tuple[int, ...], *, dtype: DType | None = None, device: Device | None = None
) -> jax.Array:
    return array_api.empty(shape, dtype=dtype, device=device)


# =============================================================================


@dispatcher  # type: ignore[misc]
def empty_like(
    x: ArrayLike, /, *, dtype: DType | None = None, device: Device | None = None
) -> ArrayLike:
    return array_api.empty_like(x, dtype=dtype, device=device)


# =============================================================================


def eye(
    n_rows: int,
    n_cols: int | None = None,
    /,
    *,
    k: int = 0,
    dtype: DType | None = None,
    device: Device | None = None,
) -> jax.Array:
    return array_api.eye(n_rows, n_cols, k=k, dtype=dtype, device=device)


# =============================================================================


def from_dlpack(x: object, /) -> jax.Array:
    return array_api.from_dlpack(x)


# =============================================================================


@dispatcher
def full(
    shape: tuple[int, ...] | int,
    fill_value: ArrayLike,
    *,
    dtype: DType | None = None,
    device: Device | None = None,
) -> ArrayLike:
    return array_api.full(shape, fill_value, dtype=dtype, device=device)


@dispatcher  # type: ignore[no-redef]
def full(
    shape: tuple[int, ...] | int,
    *,
    fill_value: ArrayLike,
    dtype: DType | None = None,
    device: Device | None = None,
) -> ArrayLike:
    return full(shape, fill_value, dtype=dtype, device=device)


# =============================================================================


@dispatcher
def full_like(
    x: ArrayLike,
    /,
    fill_value: ArrayLike,
    *,
    dtype: DType | None = None,
    device: Device | None = None,
) -> ArrayLike:
    return array_api.full_like(x, fill_value, dtype=dtype, device=device)


@dispatcher  # type: ignore[no-redef]
def full_like(
    x: ArrayLike,
    *,
    fill_value: ArrayLike,
    dtype: DType | None = None,
    device: Device | None = None,
) -> ArrayLike:
    # dispatch on both `x` and `fill_value`
    return full_like.invoke(type(x), type(fill_value))(
        x, fill_value, dtype=dtype, device=device
    )


# =============================================================================


@dispatcher
def linspace(  # noqa: PLR0913
    start: ArrayLike,
    stop: ArrayLike,
    num: int,
    /,
    *,
    dtype: DType | None = None,
    device: Device | None = None,
    endpoint: bool = True,
) -> jax.Array | jax.core.Tracer | Value:
    return array_api.linspace(
        start,
        stop,
        num,
        dtype=dtype,
        device=device,
        endpoint=endpoint,
    )


@dispatcher  # type: ignore[no-redef]
def linspace(  # noqa: PLR0913
    start: ArrayLike,
    stop: ArrayLike,
    /,
    *,
    num: int,
    dtype: DType | None = None,
    device: Device | None = None,
    endpoint: bool = True,
) -> ArrayLike:
    # dispatch on `start`, `stop`, and `num`
    return linspace(start, stop, num, dtype=dtype, device=device, endpoint=endpoint)


# =============================================================================


@quaxify
def meshgrid(*arrays: ArrayLike, indexing: str = "xy") -> list[ArrayLike]:
    return jnp.meshgrid(*arrays, indexing=indexing)


# =============================================================================


def ones(
    shape: tuple[int, ...],
    *,
    dtype: DType | None = None,
    device: Device | None = None,
) -> jax.Array:
    return array_api.ones(shape, dtype=dtype, device=device)


# =============================================================================


@dispatcher  # type: ignore[misc]
def ones_like(
    x: ArrayLike, /, *, dtype: DType | None = None, device: Device | None = None
) -> ArrayLike:
    return array_api.ones_like(x, dtype=dtype, device=device)


# =============================================================================


@quaxify
def tril(x: ArrayLike, /, *, k: int = 0) -> ArrayLike:
    return array_api.tril(x, k=k)


# =============================================================================


@quaxify
def triu(x: ArrayLike, /, *, k: int = 0) -> ArrayLike:
    return array_api.triu(x, k=k)


# =============================================================================


def zeros(
    shape: tuple[int, ...],
    *,
    dtype: DType | None = None,
    device: Device | None = None,
) -> jax.Array:
    return array_api.zeros(shape, dtype=dtype, device=device)


# =============================================================================


@dispatcher  # type: ignore[misc]
def zeros_like(
    x: ArrayLike,
    /,
    *,
    dtype: DType | None = None,
    device: Device | None = None,
) -> ArrayLike | jax.Array:
    return array_api.zeros_like(x, dtype=dtype, device=device)
